@@ -3,6 +3,7 @@ import logging
 from angr.utils.constants import DEFAULT_STATEMENT
 from angr.engines.pcode.lifter import IRSB
 from pypcode import OpCode, Varnode
+import pypcode
 
 from .block import Block
 from .statement import Statement, Assignment, Store, Jump, ConditionalJump, Return, Call
@@ -31,9 +32,9 @@ opcode_to_generic_name = {
     # OpCode.INT_SEXT          : '',
     OpCode.INT_ADD: "Add",
     OpCode.INT_SUB: "Sub",
-    # OpCode.INT_CARRY         : '',
-    # OpCode.INT_SCARRY        : '',
-    # OpCode.INT_SBORROW       : '',
+    OpCode.INT_CARRY: "Carry",
+    OpCode.INT_SCARRY: "SCarry",
+    OpCode.INT_SBORROW: "SBorrow",
     # OpCode.INT_2COMP         : '',
     OpCode.INT_NEGATE: "Neg",
     OpCode.INT_XOR: "Xor",
@@ -48,7 +49,7 @@ opcode_to_generic_name = {
     # OpCode.INT_REM           : '',
     # OpCode.INT_SREM          : '',
     OpCode.BOOL_NEGATE: "Not",
-    # OpCode.BOOL_XOR          : '',
+    OpCode.BOOL_XOR: "LogicalXor",
     OpCode.BOOL_AND: "LogicalAnd",
     OpCode.BOOL_OR: "LogicalOr",
     # OpCode.CAST              : '',
@@ -101,13 +102,13 @@ class PCodeIRSBConverter(Converter):
         self._irsb = irsb
         self._manager = manager
         self._statements = []
-        self._current_ins = None
         self._current_op = None
+        self._next_ins_addr = None
         self._current_behavior = None
         self._statement_idx = 0
 
         # Remap all uniques s.t. they are write-once with values starting from 0
-        self._unique_tracker = {}
+        self._unique_tracker: dict[int, tuple[int, int]] = {}
         self._unique_counter = 0
 
         self._special_op_handlers = {
@@ -141,14 +142,20 @@ class PCodeIRSBConverter(Converter):
         Convert the given IRSB to an AIL Block
         """
         self._statement_idx = 0
-        for ins in self._irsb._instructions:
-            self._current_ins = ins
-            self._manager.ins_addr = ins.address.offset
-            for op in self._current_ins.ops:
-                self._current_op = op
+
+        for op in self._irsb._ops:
+            self._current_op = op
+            if op.opcode == pypcode.OpCode.IMARK:
+                self._manager.ins_addr = op.inputs[0].offset
+                self._next_ins_addr = op.inputs[-1].offset + op.inputs[-1].size
+            else:
                 self._current_behavior = self._irsb.behaviors.get_behavior_for_opcode(self._current_op.opcode)
                 self._convert_current_op()
-                self._statement_idx += 1
+            self._statement_idx += 1
+
+            if "sparc:" in self._irsb.arch.name and self._irsb.arch.bits == 32:
+                if self._current_op.opcode == OpCode.CALL:
+                    break
 
         return Block(self._irsb.addr, self._irsb.size, statements=self._statements)
 
@@ -179,8 +186,8 @@ class PCodeIRSBConverter(Converter):
         op = opcode_to_generic_name.get(opcode, None)
         in1 = self._get_value(self._current_op.inputs[0])
         if op is None:
-            log.warning("p-code: Unsupported opcode of type %s", opcode.name)
-            out = DirtyExpression(self._manager.next_atom(), opcode.name, bits=self._current_op.output.size * 8)
+            log.warning("p-code: Unsupported opcode of type %s", opcode.__name__)
+            out = DirtyExpression(self._manager.next_atom(), opcode.__name__, bits=self._current_op.output.size * 8)
         else:
             out = UnaryOp(self._manager.next_atom(), op, in1, ins_addr=self._manager.ins_addr)
 
@@ -198,8 +205,8 @@ class PCodeIRSBConverter(Converter):
         signed = op in {"CmpLEs", "CmpGTs"}
 
         if op is None:
-            log.warning("p-code: Unsupported opcode of type %s.", opcode.name)
-            out = DirtyExpression(self._manager.next_atom(), opcode.name, bits=self._current_op.output.size * 8)
+            log.warning("p-code: Unsupported opcode of type %s.", opcode.__name__)
+            out = DirtyExpression(self._manager.next_atom(), opcode.__name__, bits=self._current_op.output.size * 8)
         else:
             out = BinaryOp(self._manager.next_atom(), op, [in1, in2], signed, ins_addr=self._manager.ins_addr)
 
@@ -228,7 +235,7 @@ class PCodeIRSBConverter(Converter):
         # FIXME: Will need performance optimization
         # FIXME: Should not get trans object this way. Moreover, should have a
         #        faster mapping method than going through trans
-        reg_name = varnode.get_register_name()
+        reg_name = varnode.getRegisterName()
         try:
             reg_offset = self._manager.arch.get_register_offset(reg_name.lower())
             log.debug("Mapped register '%s' to offset %x", reg_name, reg_offset)
@@ -237,7 +244,7 @@ class PCodeIRSBConverter(Converter):
             log.warning("Could not map register '%s' from archinfo. Mapping to %x", reg_name, reg_offset)
         return reg_offset
 
-    def _remap_temp(self, offset: int, is_write: bool) -> int:
+    def _remap_temp(self, offset: int, size: int, is_write: bool) -> int | None:
         """
         Remap any unique space addresses such that they are written only once
 
@@ -246,11 +253,14 @@ class PCodeIRSBConverter(Converter):
         :return:         The remapped temporary register index
         """
         if is_write:
-            self._unique_tracker[offset] = self._unique_counter
+            self._unique_tracker[offset] = self._unique_counter, size
             self._unique_counter += 1
+            return self._unique_tracker[offset][0]
         else:
-            assert offset in self._unique_tracker
-        return self._unique_tracker[offset]
+            if offset in self._unique_tracker:
+                return self._unique_tracker[offset][0]
+            # this might be a partial access of an existing temporary variable. return None for now
+            return None
 
     def _convert_varnode(self, varnode: Varnode, is_write: bool) -> Expression:
         """
@@ -267,9 +277,32 @@ class PCodeIRSBConverter(Converter):
             return Const(self._manager.next_atom(), None, varnode.offset, size)
         elif space_name == "register":
             offset = self._map_register_name(varnode)
-            return Register(self._manager.next_atom(), None, offset, size, reg_name=varnode.get_register_name())
+            return Register(self._manager.next_atom(), None, offset, size, reg_name=varnode.getRegisterName())
         elif space_name == "unique":
-            offset = self._remap_temp(varnode.offset, is_write)
+            offset = self._remap_temp(varnode.offset, varnode.size, is_write)
+            if offset is None:
+                # this might be a partial access of an existing temporary variable
+                unique_offset = None
+                for delta in range(-1, -8, -1):
+                    if varnode.offset + delta in self._unique_tracker:
+                        unique_offset = varnode.offset + delta
+                        break
+                assert unique_offset is not None, "Cannot find the source unique variable"
+                # TODO: Check size
+                _, ori_tmp_size = self._unique_tracker[unique_offset]
+                t = Tmp(self._manager.next_atom(), None, unique_offset, ori_tmp_size * 8)
+                # FIXME: Asserting BE
+                right_shift_amount = varnode.offset + varnode.size - (unique_offset + ori_tmp_size)
+                if right_shift_amount != 0:
+                    t = BinaryOp(
+                        self._manager.next_atom(),
+                        "Shr",
+                        [t, Const(self._manager.next_atom(), None, right_shift_amount * 8, 8)],
+                        False,
+                        ins_addr=self._manager.ins_addr,
+                    )
+                return Convert(self._manager.next_atom(), t.bits, size, False, t, ins_addr=self._manager.ins_addr)
+
             return Tmp(self._manager.next_atom(), None, offset, size)
         elif space_name in ["ram", "mem"]:
             assert not is_write
@@ -384,7 +417,7 @@ class PCodeIRSBConverter(Converter):
         """
         Convert a p-code load operation
         """
-        spc = self._current_op.inputs[0].get_space_from_const()
+        spc = self._current_op.inputs[0].getSpaceFromConst()
         out = self._current_op.output
         assert spc.name in {"ram", "mem", "register"}
         if spc.name == "register":
@@ -408,33 +441,40 @@ class PCodeIRSBConverter(Converter):
         """
         Convert a p-code store operation
         """
-        spc = self._current_op.inputs[0].get_space_from_const()
-        assert spc.name in ["ram", "mem"]
-        off = self._get_value(self._current_op.inputs[1])
-        data = self._get_value(self._current_op.inputs[2])
-        log.debug("Storing %s at offset %s", data, off)
-        # self.state.memory.store(off, data, endness=self.project.arch.memory_endness)
-        stmt = Store(
-            self._statement_idx,
-            off,
-            data,
-            self._current_op.inputs[2].size,
-            self._manager.arch.memory_endness,
-            ins_addr=self._manager.ins_addr,
-        )
+        spc = self._current_op.inputs[0].getSpaceFromConst()
+        assert spc.name in {"ram", "mem", "register"}
+        if spc.name == "register":
+            # store to register
+            out = self._current_op.inputs[2]
+            res = self._get_value(self._current_op.inputs[1])
+            stmt = self._set_value(out, res)
+        else:
+            # store to memory
+            off = self._get_value(self._current_op.inputs[1])
+            data = self._get_value(self._current_op.inputs[2])
+            log.debug("Storing %s at offset %s", data, off)
+            # self.state.memory.store(off, data, endness=self.project.arch.memory_endness)
+            stmt = Store(
+                self._statement_idx,
+                off,
+                data,
+                self._current_op.inputs[2].size,
+                self._manager.arch.memory_endness,
+                ins_addr=self._manager.ins_addr,
+            )
         self._statements.append(stmt)
 
     def _convert_branch(self) -> None:
         """
         Convert a p-code branch operation
         """
-        dest_addr = self._current_op.inputs[0].get_addr()
-        if dest_addr.is_constant:
+        if self._current_op.inputs[0].space == "const":
             raise NotImplementedError("p-code relative branch not supported yet")
+        dest_addr = self._current_op.inputs[0].offset
 
         # special handling: if the previous statement is a ConditionalJump with a None destination address, then we
         # back-patch the previous statement
-        dest = Const(self._manager.next_atom(), None, dest_addr.offset, self._manager.arch.bits)
+        dest = Const(self._manager.next_atom(), None, dest_addr, self._manager.arch.bits)
         if self._statements:
             last_stmt = self._statements[-1]
             if isinstance(last_stmt, ConditionalJump) and last_stmt.false_target is None:
@@ -448,20 +488,19 @@ class PCodeIRSBConverter(Converter):
         """
         Convert a p-code conditional branch operation
         """
-        cond = self._get_value(self._current_op.inputs[1])
-        dest_addr = self._current_op.inputs[0].get_addr()
-        if dest_addr.is_constant:
+        if self._current_op.inputs[0].space == "const":
             raise NotImplementedError("p-code relative branch not supported yet")
-        dest_addr = dest_addr.offset
+        dest_addr = self._current_op.inputs[0].offset
+        cond = self._get_value(self._current_op.inputs[1])
         cval = Const(self._manager.next_atom(), None, 0, cond.bits)
         condition = BinaryOp(self._manager.next_atom(), "CmpNE", [cond, cval], signed=False)
         dest = Const(self._manager.next_atom(), None, dest_addr, self._manager.arch.bits)
-        if self._current_ins.ops[-1] is self._current_op:
+        if self._irsb._ops[-1] is self._current_op:
             # if the cbranch op is the last op, then we need to generate a fallthru target
             fallthru = Const(
                 self._manager.next_atom(),
                 None,
-                self._manager.ins_addr + self._current_ins.length,
+                self._next_ins_addr,
                 self._manager.arch.bits,
             )
         else:
