@@ -1,5 +1,8 @@
 # pylint:disable=arguments-renamed,isinstance-second-argument-not-valid-type,missing-class-docstring
-from typing import Optional, List, TYPE_CHECKING
+from __future__ import annotations
+from enum import IntEnum
+
+from typing import TYPE_CHECKING
 
 try:
     import claripy
@@ -65,7 +68,7 @@ class Atom(Expression):
         "variable_offset",
     )
 
-    def __init__(self, idx, variable, variable_offset=0, **kwargs):
+    def __init__(self, idx, variable=None, variable_offset=0, **kwargs):
         super().__init__(idx, 0, **kwargs)
         self.variable = variable
         self.variable_offset = variable_offset
@@ -97,13 +100,20 @@ class Const(Atom):
         return str(self)
 
     def __str__(self):
-        try:
+        if isinstance(self.value, int):
             return "%#x<%d>" % (self.value, self.bits)
-        except TypeError:
+        elif isinstance(self.value, float):
             return "%f<%d>" % (self.value, self.bits)
+        else:
+            return f"{self.value}<{self.bits}>"
 
     def likes(self, other):
-        return type(self) is type(other) and self.value == other.value and self.bits == other.bits
+        # nan is nan, but nan != nan
+        return (
+            type(self) is type(other)
+            and (self.value is other.value or self.value == other.value)
+            and self.bits == other.bits
+        )
 
     __hash__ = TaggedObject.__hash__
 
@@ -114,7 +124,7 @@ class Const(Atom):
     def sign_bit(self):
         return self.value >> (self.bits - 1)
 
-    def copy(self) -> "Const":
+    def copy(self) -> Const:
         return Const(self.idx, self.variable, self.value, self.bits, **self.tags)
 
 
@@ -148,7 +158,7 @@ class Tmp(Atom):
     def _hash_core(self):
         return stable_hash(("tmp", self.tmp_idx, self.bits))
 
-    def copy(self) -> "Tmp":
+    def copy(self) -> Tmp:
         return Tmp(self.idx, self.variable, self.tmp_idx, self.bits, **self.tags)
 
 
@@ -187,10 +197,181 @@ class Register(Atom):
     def _hash_core(self):
         return stable_hash(("reg", self.reg_offset, self.bits, self.idx))
 
-    def copy(self) -> "Register":
+    def copy(self) -> Register:
         return Register(self.idx, self.variable, self.reg_offset, self.bits, **self.tags)
 
 
+class VirtualVariableCategory(IntEnum):
+    REGISTER = 0
+    STACK = 1
+    MEMORY = 2
+    UNKNOWN = 3
+
+
+class VirtualVariable(Atom):
+
+    __slots__ = (
+        "bits",
+        "varid",
+        "category",
+        "oident",
+    )
+
+    def __init__(
+        self,
+        idx,
+        varid: int,
+        bits,
+        category: VirtualVariableCategory,
+        oident: int | str | None = None,
+        **kwargs,
+    ):
+        super().__init__(idx, **kwargs)
+
+        self.varid = varid
+        self.category = category
+        self.oident = oident
+        self.bits = bits
+
+    @property
+    def size(self):
+        return self.bits // 8
+
+    @property
+    def was_reg(self) -> bool:
+        return self.category == VirtualVariableCategory.REGISTER
+
+    @property
+    def was_stack(self) -> bool:
+        return self.category == VirtualVariableCategory.STACK
+
+    @property
+    def reg_offset(self) -> int | None:
+        if self.was_reg:
+            return self.oident
+        return None
+
+    @property
+    def stack_offset(self) -> int | None:
+        if self.was_stack:
+            return self.oident
+        return None
+
+    def likes(self, atom):
+        return (
+            isinstance(atom, VirtualVariable)
+            and self.varid == atom.varid
+            and self.bits == atom.bits
+            and self.category == atom.category
+            and self.oident == atom.oident
+        )
+
+    def __repr__(self):
+        ori_str = ""
+        match self.category:
+            case VirtualVariableCategory.REGISTER:
+                ori_str = f"{{reg {self.reg_offset}}}"
+            case VirtualVariableCategory.STACK:
+                ori_str = f"{{stack {self.oident}}}"
+        return f"vvar_{self.varid}{ori_str}"
+
+    __hash__ = TaggedObject.__hash__
+
+    def _hash_core(self):
+        return stable_hash(("var", self.varid, self.bits, self.category, self.oident))
+
+    def copy(self) -> VirtualVariable:
+        return VirtualVariable(
+            self.idx,
+            self.varid,
+            self.bits,
+            self.category,
+            oident=self.oident,
+            variable=self.variable,
+            variable_offset=self.variable_offset,
+            **self.tags,
+        )
+
+
+class Phi(Atom):
+
+    __slots__ = (
+        "bits",
+        "src_and_vvars",
+    )
+
+    def __init__(
+        self,
+        idx,
+        bits,
+        src_and_vvars: list[tuple[tuple[int, int], VirtualVariable]],
+        **kwargs,
+    ):
+        super().__init__(idx, **kwargs)
+        self.bits = bits
+        self.src_and_vvars = src_and_vvars
+
+    @property
+    def size(self) -> int:
+        return self.bits // 8
+
+    @property
+    def op(self) -> str:
+        return "Phi"
+
+    @property
+    def verbose_op(self) -> str:
+        return "Phi"
+
+    def likes(self, atom) -> bool:
+        if isinstance(atom, Phi) and self.bits == atom.bits:
+            self_src_and_vvarids = {(src, vvar.varid if vvar is not None else None) for src, vvar in self.src_and_vvars}
+            other_src_and_vvarids = {
+                (src, vvar.varid if vvar is not None else None) for src, vvar in atom.src_and_vvars
+            }
+            return self_src_and_vvarids == other_src_and_vvarids
+        return False
+
+    def __repr__(self):
+        return f"𝜙@{self.bits}b {self.src_and_vvars}"
+
+    __hash__ = TaggedObject.__hash__
+
+    def _hash_core(self):
+        return stable_hash(("phi", self.bits, tuple(sorted(self.src_and_vvars))))
+
+    def copy(self) -> Phi:
+        return Phi(
+            self.idx,
+            self.bits,
+            self.src_and_vvars[::],
+            variable=self.variable,
+            variable_offset=self.variable_offset,
+            **self.tags,
+        )
+
+    def replace(self, old_expr, new_expr):
+        replaced = False
+        new_src_and_vvars = []
+        for src, vvar in self.src_and_vvars:
+            if vvar == old_expr and isinstance(new_expr, VirtualVariable):
+                replaced = True
+                new_src_and_vvars.append((src, new_expr))
+            else:
+                new_src_and_vvars.append((src, vvar))
+
+        if replaced:
+            return True, Phi(
+                self.idx,
+                self.bits,
+                new_src_and_vvars,
+                variable=self.variable,
+                variable_offset=self.variable_offset,
+                **self.tags,
+            )
+        return False, self
+
+
 class Op(Expression):
     __slots__ = ("op",)
 
@@ -255,7 +436,7 @@ class UnaryOp(Op):
     def size(self):
         return self.bits // 8
 
-    def copy(self) -> "UnaryOp":
+    def copy(self) -> UnaryOp:
         return UnaryOp(
             self.idx, self.op, self.operand, variable=self.variable, variable_offset=self.variable_offset, **self.tags
         )
@@ -369,7 +550,7 @@ class Convert(UnaryOp):
         else:
             return False, self
 
-    def copy(self) -> "Convert":
+    def copy(self) -> Convert:
         return Convert(
             self.idx,
             self.from_bits,
@@ -446,7 +627,7 @@ class Reinterpret(UnaryOp):
         else:
             return False, self
 
-    def copy(self) -> "Reinterpret":
+    def copy(self) -> Reinterpret:
         return Reinterpret(
             self.idx, self.from_bits, self.from_type, self.to_bits, self.to_type, self.operand, **self.tags
         )
@@ -461,15 +642,19 @@ class BinaryOp(Op):
         "variable_offset",
         "floating_point",
         "rounding_mode",
+        "from_bits",  # for divmod
+        "to_bits",  # for divmod
     )
 
     OPSTR_MAP = {
         "Add": "+",
         "AddF": "+",
+        "AddV": "+",
         "Sub": "-",
         "SubF": "-",
         "Mul": "*",
         "MulF": "*",
+        "MulV": "*",
         "Div": "/",
         "DivF": "/",
         "DivMod": "/m",
@@ -496,6 +681,9 @@ class BinaryOp(Op):
         "Concat": "CONCAT",
         "Ror": "ROR",
         "Rol": "ROL",
+        "Carry": "CARRY",
+        "SCarry": "SCARRY",
+        "SBorrow": "SBORROW",
     }
 
     COMPARISON_NEGATION = {
@@ -522,6 +710,8 @@ class BinaryOp(Op):
         bits=None,
         floating_point=False,
         rounding_mode=None,
+        from_bits=None,
+        to_bits=None,
         **kwargs,
     ):
         depth = (
@@ -546,8 +736,18 @@ class BinaryOp(Op):
             self.bits = bits
         elif self.op == "CmpF":
             self.bits = 32  # floating point comparison
-        elif self.op.startswith("Cmp"):
+        elif self.op in {
+            "CmpEQ",
+            "CmpNE",
+            "CmpLT",
+            "CmpGE",
+            "CmpLE",
+            "CmpGT",
+            "ExpCmpNE",
+        }:
             self.bits = 1
+        elif self.op in {"Carry", "SCarry", "SBorrow"}:
+            self.bits = 8
         elif self.op == "Concat":
             self.bits = get_bits(operands[0]) + get_bits(operands[1])
         elif self.op == "Mull":
@@ -560,6 +760,9 @@ class BinaryOp(Op):
         self.floating_point = floating_point
         self.rounding_mode = rounding_mode
 
+        self.from_bits = from_bits
+        self.to_bits = to_bits
+
         # TODO: sanity check of operands' sizes for some ops
         # assert self.bits == operands[1].bits
 
@@ -645,6 +848,8 @@ class BinaryOp(Op):
                 bits=self.bits,
                 floating_point=self.floating_point,
                 rounding_mode=self.rounding_mode,
+                from_bits=self.from_bits,
+                to_bits=self.to_bits,
                 **self.tags,
             )
         else:
@@ -664,7 +869,7 @@ class BinaryOp(Op):
     def size(self):
         return self.bits // 8
 
-    def copy(self) -> "BinaryOp":
+    def copy(self) -> BinaryOp:
         return BinaryOp(
             self.idx,
             self.op,
@@ -675,6 +880,8 @@ class BinaryOp(Op):
             bits=self.bits,
             floating_point=self.floating_point,
             rounding_mode=self.rounding_mode,
+            from_bits=self.from_bits,
+            to_bits=self.to_bits,
             **self.tags,
         )
 
@@ -778,7 +985,7 @@ class TernaryOp(Op):
     def size(self):
         return self.bits // 8
 
-    def copy(self) -> "TernaryOp":
+    def copy(self) -> TernaryOp:
         return TernaryOp(self.idx, self.op, self.operands[::], bits=self.bits, **self.tags)
 
 
@@ -856,7 +1063,7 @@ class Load(Expression):
     def _hash_core(self):
         return stable_hash(("Load", self.addr, self.size, self.endness))
 
-    def copy(self) -> "Load":
+    def copy(self) -> Load:
         return Load(
             self.idx,
             self.addr,
@@ -958,7 +1165,7 @@ class ITE(Expression):
     def size(self):
         return self.bits // 8
 
-    def copy(self) -> "ITE":
+    def copy(self) -> ITE:
         return ITE(self.idx, self.cond, self.iffalse, self.iftrue, **self.tags)
 
 
@@ -987,7 +1194,7 @@ class DirtyExpression(Expression):
     def __str__(self):
         return "[D] %s" % str(self.dirty_expr)
 
-    def copy(self) -> "DirtyExpression":
+    def copy(self) -> DirtyExpression:
         return DirtyExpression(self.idx, self.dirty_expr, bits=self.bits, **self.tags)
 
     def replace(self, old_expr, new_expr):
@@ -1043,7 +1250,7 @@ class VEXCCallExpression(Expression):
         operands_str = ", ".join(repr(op) for op in self.operands)
         return f"{self.cee_name}({operands_str})"
 
-    def copy(self) -> "VEXCCallExpression":
+    def copy(self) -> VEXCCallExpression:
         return VEXCCallExpression(self.idx, self.cee_name, self.operands, bits=self.bits, **self.tags)
 
     def replace(self, old_expr, new_expr):
@@ -1081,7 +1288,7 @@ class MultiStatementExpression(Expression):
         "expr",
     )
 
-    def __init__(self, idx: Optional[int], stmts: List["Statement"], expr: Expression, **kwargs):
+    def __init__(self, idx: int | None, stmts: list[Statement], expr: Expression, **kwargs):
         super().__init__(idx, expr.depth + 1, **kwargs)
         self.stmts = stmts
         self.expr = expr
@@ -1133,7 +1340,7 @@ class MultiStatementExpression(Expression):
             )
         return False, self
 
-    def copy(self) -> "MultiStatementExpression":
+    def copy(self) -> MultiStatementExpression:
         return MultiStatementExpression(self.idx, self.stmts[::], self.expr, **self.tags)
 
 
@@ -1204,7 +1411,7 @@ class BasePointerOffset(Expression):
             return True, BasePointerOffset(self.idx, self.bits, new_base, new_offset, **self.tags)
         return False, self
 
-    def copy(self) -> "BasePointerOffset":
+    def copy(self) -> BasePointerOffset:
         return BasePointerOffset(self.idx, self.bits, self.base, self.offset, **self.tags)
 
 
@@ -1217,7 +1424,7 @@ class StackBaseOffset(BasePointerOffset):
             offset -= 1 << bits
         super().__init__(idx, bits, "stack_base", offset, **kwargs)
 
-    def copy(self) -> "StackBaseOffset":
+    def copy(self) -> StackBaseOffset:
         return StackBaseOffset(self.idx, self.bits, self.offset, **self.tags)
 
 
