@@ -224,7 +224,7 @@ class Jump(Statement):
         "target_idx",
     )
 
-    def __init__(self, idx, target, target_idx: Optional[int] = None, **kwargs):
+    def __init__(self, idx, target, target_idx: int | None = None, **kwargs):
         super().__init__(idx, **kwargs)
 
         self.target = target
@@ -251,6 +251,10 @@ class Jump(Statement):
             return f"Goto({self.target}.{self.target_idx})"
         return "Goto(%s)" % self.target
 
+    @property
+    def depth(self):
+        return self.target.depth
+
     def replace(self, old_expr, new_expr):
         r, replaced_target = self.target.replace(old_expr, new_expr)
 
@@ -286,8 +290,8 @@ class ConditionalJump(Statement):
         condition,
         true_target,
         false_target,
-        true_target_idx: Optional[int] = None,
-        false_target_idx: Optional[int] = None,
+        true_target_idx: int | None = None,
+        false_target_idx: int | None = None,
         **kwargs,
     ):
         super().__init__(idx, **kwargs)
@@ -315,8 +319,6 @@ class ConditionalJump(Statement):
             and self.condition.likes(other.condition)
             and is_none_or_likeable(self.true_target, other.true_target)
             and is_none_or_likeable(self.false_target, other.false_target)
-            and self.true_target == other.true_target
-            and self.false_target == other.false_target
         )
 
     __hash__ = TaggedObject.__hash__
@@ -667,7 +669,7 @@ class Label(Statement):
         "block_idx",
     )
 
-    def __init__(self, idx, name: str, ins_addr: int, block_idx: Optional[int] = None, **kwargs):
+    def __init__(self, idx, name: str, ins_addr: int, block_idx: int | None = None, **kwargs):
         super().__init__(idx, **kwargs)
         self.name = name
         self.ins_addr = ins_addr
